use std::{alloc::Layout, ffi::{c_char, CStr}, path::PathBuf};

use dlopen::raw::Library;

#[derive(Debug, Clone, Copy,)]
pub struct PluginSyms {
    init:        unsafe extern "C" fn(),
    pre_reload:  unsafe extern "C" fn() -> *const (),
    post_reload: unsafe extern "C" fn(state: *const ()),
    poll:        unsafe extern "C" fn(buf: *mut i8, len: usize),
    free:        unsafe extern "C" fn(),
    get_info:    unsafe extern "C" fn() -> *const PluginInfo, 
}

#[repr(C)]
pub struct PluginInfo {
    name: *const c_char,
    version: *const c_char,
    license: *const c_char,
}

#[derive(Debug, Default)]
pub struct Plugin {
    syms: Option<PluginSyms>,
    path: PathBuf,
    lib: Option<Library>,
    enabled: bool,
}


impl Plugin {
    pub fn load(path: PathBuf) -> Result<Self, dlopen::Error> {
        let mut s = Self::default();
        s.enabled = true;
        s.path = path;
        s.reload_symbols()?;
        Ok(s)
    }

    fn reload_symbols(&mut self) -> Result<(), dlopen::Error> {
        log::debug!("Loading {:?}", self.path);
        let lib = Library::open(&self.path)?;
        let symbols = PluginSyms { 
            init: unsafe { lib.symbol("plug_init")? },
            pre_reload: unsafe { lib.symbol("plug_pre_reload")? },
            post_reload: unsafe { lib.symbol("plug_post_reload")? },
            poll: unsafe { lib.symbol("plug_poll")? },
            free: unsafe { lib.symbol("plug_free")? },

            get_info: unsafe { lib.symbol("plug_get_info")? },
        };
        unsafe {
            if (symbols.get_info)().is_null() {
                log::error!("Info fields for plugin {:?} are null", self.path);
                self.disable();
            }
        }

        self.syms = Some(symbols);
        self.lib = Some(lib);
        Ok(())
    }
    pub fn syms(&self) -> PluginSyms {
        self.syms.unwrap()
    }

    pub fn init(&self) {
        unsafe {
            (self.syms().init)()
        }
    }
    pub fn reload(&mut self) -> Result<(), dlopen::Error> { 
        unsafe {
            let state = (self.syms().pre_reload)();
            if let Err(e) = self.reload_symbols() {
                self.disable();
                return Err(e);
            }
            (self.syms().post_reload)(state);
        }
        Ok(())
    }

    pub fn poll(&self, cap: usize) -> Result<String, std::string::FromUtf8Error> {
        let layout = Layout::from_size_align(cap, 1).unwrap();
        let buf = unsafe {
            std::alloc::alloc_zeroed(layout)
        };
        let s = unsafe {
            (self.syms().poll)(buf as *mut i8, cap);
            let len = libc::strlen(buf as *const i8);
            if len > cap {
                panic!("String len is bigger than allocatd");
            }

            String::from_raw_parts(buf, len, cap)
        };

        if !s.is_empty() {
            log::debug!("Polled: {}", s);
        }
        Ok(s)
    }

    pub fn free(self) {
        unsafe {
            (self.syms().free)();
        }
        drop(self.lib.unwrap()); 
    }

    pub fn enabled(&self) -> bool {
        self.enabled
    }

    pub fn disable(&mut self) {
        self.enabled = false;
    }

    pub fn name(&self) -> String {
        unsafe {
            let info = (self.syms().get_info)();
            CStr::from_ptr((*info).name).to_string_lossy().to_string()
        }
    }

    pub fn version(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).version;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }

    pub fn license(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).license;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }
}

